import functools
import warnings
from typing import Any, List, Optional
from tenacity import (
    Retrying,
    stop_after_attempt,
    wait_exponential_jitter,
    retry_if_exception_type,
)

import voyageai
import voyageai.error as error
from voyageai.util import default_api_key
from voyageai.object import EmbeddingsObject, RerankingObject


class Client:
    """Voyage AI Client

    Args:
        api_key (str): Your API key.
        max_retries (int): Maximum number of retries if API call fails.
        timeout (float): Timeout in seconds.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        max_retries: int = 0,
        timeout: Optional[float] = None,
    ) -> None:

        self.api_key = api_key or default_api_key()

        self._params = {
            "api_key": self.api_key,
            "request_timeout": timeout,
        }

        self.retry_controller = Retrying(
            reraise=True,
            stop=stop_after_attempt(max_retries),
            wait=wait_exponential_jitter(initial=1, max=16),
            retry=(
                retry_if_exception_type(error.RateLimitError)
                | retry_if_exception_type(error.ServiceUnavailableError)
                | retry_if_exception_type(error.Timeout)
            ),
        )

    def embed(
        self,
        texts: List[str],
        model: Optional[str] = None,
        input_type: Optional[str] = None,
        truncation: bool = True,
    ) -> EmbeddingsObject:

        if model is None:
            model = voyageai.VOYAGE_EMBED_DEFAULT_MODEL
            warnings.warn(
                f"The `model` argument is not specified and defaults to {voyageai.VOYAGE_EMBED_DEFAULT_MODEL}. "
                "It will be a required argument in the future. We recommend to specify the model when using this "
                "function. Please see https://docs.voyageai.com/docs/embeddings for the list of latest models "
                "provided by Voyage AI."
            )

        for attempt in self.retry_controller:
            with attempt:
                response = voyageai.Embedding.create(
                    input=texts,
                    model=model,
                    input_type=input_type,
                    truncation=truncation,
                    **self._params,
                )

        result = EmbeddingsObject(response)
        return result

    def rerank(
        self,
        query: str,
        documents: List[str],
        model: str,
        top_k: Optional[int] = None,
        truncation: bool = True,
    ) -> RerankingObject:

        for attempt in self.retry_controller:
            with attempt:
                response = voyageai.Reranking.create(
                    query=query,
                    documents=documents,
                    model=model,
                    top_k=top_k,
                    truncation=truncation,
                    **self._params,
                )

        result = RerankingObject(documents, response)
        return result

    @functools.lru_cache()
    def tokenizer(self, model: str):
        try:
            from tokenizers import Tokenizer
        except ImportError:
            raise ImportError(
                "The package `tokenizers` is not found. Please run `pip install tokenizers` "
                "to install the dependency."
            )

        try:
            tokenizer = Tokenizer.from_pretrained(f"voyageai/{model}")
            tokenizer.no_truncation()
        except:
            warnings.warn(
                f"Failed to load the tokenizer for `{model}`. Please ensure that it is a valid model name."
            )
            raise
        
        return tokenizer

    def tokenize(
        self,
        texts: List[str],
        model: Optional[str] = None,
    ) -> List[Any]:

        if model is None:
            warnings.warn(
                "Please specify the `model` when using the tokenizer. Voyage's older models use the same "
                "tokenizer, but new models may use different tokenizers. If `model` is not specified, "
                "the old tokenizer will be used and the results might be different. `model` will be a "
                "required argument in the future."
            )
            model = voyageai.VOYAGE_EMBED_DEFAULT_MODEL

        return self.tokenizer(model).encode_batch(texts)

    def count_tokens(
        self,
        texts: List[str],
        model: Optional[str] = None,
    ) -> int:
        tokenized = self.tokenize(texts, model)
        return sum([len(t) for t in tokenized])



from collections import namedtuple
from typing import List, Optional
from voyageai.api_resources import VoyageResponse


RerankingResult = namedtuple(
    "RerankingResult", ["index", "document", "relevance_score"]
)


class RerankingObject:

    def __init__(self, documents: List[str], response: VoyageResponse):
        self.results: List[RerankingResult] = [
            RerankingResult(
                index=d.index,
                document=documents[d.index],
                relevance_score=d.relevance_score,
            )
            for d in response.data
        ]
        self.total_tokens: int = response.usage.total_tokens



import json
from copy import deepcopy

from voyageai.api_resources.api_requestor import VoyageHttpResponse


class VoyageResponse(dict):

    def __init__(
        self,
        **params,
    ):
        super(VoyageResponse, self).__init__()

    def __setattr__(self, k, v):
        if k[0] == "_" or k in self.__dict__:
            return super(VoyageResponse, self).__setattr__(k, v)

        self[k] = v
        return None

    def __getattr__(self, k):
        if k[0] == "_":
            raise AttributeError(k)
        try:
            return self[k]
        except KeyError as err:
            raise AttributeError(*err.args)

    def __delattr__(self, k):
        if k[0] == "_" or k in self.__dict__:
            return super(VoyageResponse, self).__delattr__(k)
        else:
            del self[k]

    def __setitem__(self, k, v):
        if v == "":
            raise ValueError(
                "You cannot set %s to an empty string. "
                "We interpret empty strings as None in requests."
                "You may set %s.%s = None to delete the property" % (k, str(self), k)
            )
        super(VoyageResponse, self).__setitem__(k, v)

    def __delitem__(self, k):
        raise NotImplementedError("del is not supported")

    # Custom unpickling method that uses `update` to update the dictionary
    # without calling __setitem__, which would fail if any value is an empty
    # string
    def __setstate__(self, state):
        self.update(state)

    # Custom pickling method to ensure the instance is pickled as a custom
    # class and not as a dict, otherwise __setstate__ would not be called when
    # unpickling.
    def __reduce__(self):
        reduce_value = (
            type(self),  # callable
            (),  # args
            dict(self),  # state
        )
        return reduce_value

    @classmethod
    def construct_from(
        cls,
        values,
    ):
        instance = cls()
        instance.refresh_from(values)
        return instance

    def refresh_from(
        self,
        values,
    ):
        # Wipe old state before setting new.
        self.clear()
        for k, v in values.items():
            super(VoyageResponse, self).__setitem__(k, convert_to_voyage_response(v))

        self._previous = values

    def __repr__(self):
        ident_parts = [type(self).__name__]

        obj = self.get("object")
        if isinstance(obj, str):
            ident_parts.append(obj)

        unicode_repr = "<%s at %s> JSON: %s" % (
            " ".join(ident_parts),
            hex(id(self)),
            str(self),
        )

        return unicode_repr

    def __str__(self):
        obj = self.to_dict_recursive()
        return json.dumps(obj, indent=2)

    def to_dict(self):
        return dict(self)

    def to_dict_recursive(self):
        d = dict(self)
        for k, v in d.items():
            if isinstance(v, VoyageResponse):
                d[k] = v.to_dict_recursive()
            elif isinstance(v, list):
                d[k] = [
                    e.to_dict_recursive() if isinstance(e, VoyageResponse) else e
                    for e in v
                ]
        return d

    # This class overrides __setitem__ to throw exceptions on inputs that it
    # doesn't like. This can cause problems when we try to copy an object
    # wholesale because some data that's returned from the API may not be valid
    # if it was set to be set manually. Here we override the class' copy
    # arguments so that we can bypass these possible exceptions on __setitem__.
    def __copy__(self):
        copied = VoyageResponse()

        for k, v in self.items():
            # Call parent's __setitem__ to avoid checks that we've added in the
            # overridden version that can throw exceptions.
            super(VoyageResponse, copied).__setitem__(k, v)

        return copied

    # This class overrides __setitem__ to throw exceptions on inputs that it
    # doesn't like. This can cause problems when we try to copy an object
    # wholesale because some data that's returned from the API may not be valid
    # if it was set to be set manually. Here we override the class' copy
    # arguments so that we can bypass these possible exceptions on __setitem__.
    def __deepcopy__(self, memo):
        copied = self.__copy__()
        memo[id(self)] = copied

        for k, v in self.items():
            # Call parent's __setitem__ to avoid checks that we've added in the
            # overridden version that can throw exceptions.
            super(VoyageResponse, copied).__setitem__(k, deepcopy(v, memo))

        return copied


def convert_to_voyage_response(resp):
    # If we get a VoyageHttpResponse, we'll want to return a VoyageResponse.

    if isinstance(resp, VoyageHttpResponse):
        resp = resp.data

    if isinstance(resp, list):
        return [convert_to_voyage_response(i) for i in resp]
    elif isinstance(resp, dict) and not isinstance(resp, VoyageResponse):
        resp = resp.copy()
        return VoyageResponse.construct_from(resp)
    else:
        return resp


def convert_to_dict(obj):
    """Converts a VoyageResponse back to a regular dict.

    Nested VoyageResponse are also converted back to regular dicts.

    :param obj: The VoyageResponse to convert.

    :returns: The VoyageResponse as a dict.
    """
    if isinstance(obj, list):
        return [convert_to_dict(i) for i in obj]
    elif isinstance(obj, dict):
        return {k: convert_to_dict(v) for k, v in obj.items()}
    else:
        return obj



from voyageai.api_resources import api_requestor
from voyageai.api_resources.response import VoyageResponse, convert_to_voyage_response


class APIResource(VoyageResponse):

    @classmethod
    def class_url(cls):
        if cls == APIResource:
            raise NotImplementedError(
                "APIResource is an abstract class. You should perform actions on its subclasses."
            )
        # Namespaces are separated in object names with periods (.) and in URLs
        # with forward slashes (/), so replace the former with the latter.
        base = cls.OBJECT_NAME.replace(".", "/")  # type: ignore
        return "/%s" % (base)

    @classmethod
    def __prepare_create_request(
        cls,
        api_key=None,
        api_base=None,
        **params,
    ):
        requestor = api_requestor.APIRequestor(
            api_key,
            api_base=api_base,
        )
        url = cls.class_url()
        headers = params.pop("headers", None)

        return requestor, url, params, headers

    @classmethod
    def create(
        cls,
        api_key=None,
        api_base=None,
        request_id=None,
        request_timeout=None,
        **params,
    ):
        requestor, url, params, headers = cls.__prepare_create_request(
            api_key, api_base, **params
        )

        response = requestor.request(
            "post",
            url,
            params=params,
            headers=headers,
            request_id=request_id,
            request_timeout=request_timeout,
        )

        obj = convert_to_voyage_response(response)
        return obj

    @classmethod
    async def acreate(
        cls,
        api_key=None,
        api_base=None,
        request_id=None,
        request_timeout=None,
        **params,
    ):
        requestor, url, params, headers = cls.__prepare_create_request(
            api_key, api_base, **params
        )
        response = await requestor.arequest(
            "post",
            url,
            params=params,
            headers=headers,
            request_id=request_id,
            request_timeout=request_timeout,
        )

        obj = convert_to_voyage_response(response)
        return obj



import asyncio
from aiolimiter import AsyncLimiter
import warnings
from typing import List, Optional
from contextlib import AsyncExitStack
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
    retry_if_exception_type,
)
from concurrent.futures import ThreadPoolExecutor

import voyageai


MAX_BATCH_SIZE = voyageai.VOYAGE_EMBED_BATCH_SIZE
MAX_LIST_LENGTH = voyageai.VOYAGE_EMBED_BATCH_SIZE
DEFAULT_CONCURRENCE = 1
DEFAULT_RPM = 300


@retry(
    wait=wait_random_exponential(min=1, max=20),
    stop=stop_after_attempt(6),
    retry=retry_if_exception_type(voyageai.error.RateLimitError),
)
def _get_embeddings(
    list_of_text: List[str],
    model: str = voyageai.VOYAGE_EMBED_DEFAULT_MODEL,
    input_type: Optional[str] = None,
    **kwargs,
) -> List[List[float]]:
    """Python wrapper for one Voyage API call."""
    _check_input_type(input_type)
    assert (
        len(list_of_text) <= MAX_BATCH_SIZE
    ), f"The length of list_of_text should not be larger than {MAX_BATCH_SIZE}."

    data = voyageai.Embedding.create(
        input=list_of_text, model=model, input_type=input_type, **kwargs
    ).data
    return [d["embedding"] for d in data]


@retry(
    wait=wait_random_exponential(min=1, max=20),
    stop=stop_after_attempt(6),
    retry=retry_if_exception_type(voyageai.error.RateLimitError),
)
async def _aget_embeddings(
    list_of_text: List[str],
    model: str = voyageai.VOYAGE_EMBED_DEFAULT_MODEL,
    input_type: Optional[str] = None,
    **kwargs,
) -> List[List[float]]:
    """Python wrapper for one async Voyage API call."""
    _check_input_type(input_type)
    assert (
        len(list_of_text) <= MAX_BATCH_SIZE
    ), f"The length of list_of_text should not be larger than {MAX_BATCH_SIZE}."

    semaphore = kwargs.pop("semaphore", AsyncExitStack())
    rate_limit = kwargs.pop("rate_limit", AsyncExitStack())

    async with semaphore:
        async with rate_limit:
            data = (
                await voyageai.Embedding.acreate(
                    input=list_of_text, model=model, **kwargs
                )
            ).data

    return [d["embedding"] for d in data]


def _check_input_type(input_type: str):
    if input_type and input_type not in ["query", "document"]:
        raise ValueError(
            f"input_type {input_type} is invalid. Options: None, 'query', 'document'."
        )


def get_embedding(
    text: str,
    model: str = voyageai.VOYAGE_EMBED_DEFAULT_MODEL,
    input_type: Optional[str] = None,
    **kwargs,
) -> List[float]:
    """Get Voyage embedding for a text string.

    Args:
        text (str): A text string to be embed.
        model (str): Name of the model to use.
        input_type (str): Type of the input text. Defalut to None, meaning the type is unspecified.
            Other options include: "query", "document".
    """
    warnings.warn(
        "This function will be deprecated. We recommend using the `embed()` function in `voyageai.Client`."
    )
    return _get_embeddings([text], model, input_type, **kwargs)[0]


def get_embeddings(
    list_of_text: List[str],
    model: str = voyageai.VOYAGE_EMBED_DEFAULT_MODEL,
    input_type: Optional[str] = None,
    **kwargs,
) -> List[List[float]]:
    """Get Voyage embedding for a list of text strings.

    Args:
        list_of_text (list): A list of text strings to embed.
        model (str): Name of the model to use.
        input_type (str): Type of the input text. Defalut to None, meaning the type is unspecified.
            Other options include: "query", "document".
    """
    warnings.warn(
        "This function will be deprecated. We recommend using the `embed()` function in `voyageai.Client`."
    )
    if len(list_of_text) <= MAX_BATCH_SIZE:
        return _get_embeddings(list_of_text, model, input_type, **kwargs)

    assert (
        len(list_of_text) <= MAX_LIST_LENGTH
    ), f"The length of list_of_text should not be larger than {MAX_LIST_LENGTH}."

    batches = [
        list_of_text[i : i + MAX_BATCH_SIZE]
        for i in range(0, len(list_of_text), MAX_BATCH_SIZE)
    ]

    with ThreadPoolExecutor(max_workers=DEFAULT_CONCURRENCE) as executor:
        futures = [
            executor.submit(_get_embeddings, batch, model, input_type)
            for batch in batches
        ]

    results = [future.result() for future in futures]
    return sum(results, [])


async def aget_embedding(
    text: str,
    model: str = voyageai.VOYAGE_EMBED_DEFAULT_MODEL,
    input_type: Optional[str] = None,
    **kwargs,
) -> List[float]:
    """Get Voyage embedding for a text string (async).

    Args:
        text (str): A text string to be embed.
        model (str): Name of the model to use.
        input_type (str): Type of the input text. Defalut to None, meaning the type is unspecified.
            Other options include: "query", "document".
    """
    warnings.warn(
        "This function will be deprecated. We recommend using the `embed()` function in `voyageai.AsyncClient`."
    )
    return (await _aget_embeddings([text], model, input_type, **kwargs))[0]


async def aget_embeddings(
    list_of_text: List[str],
    model: str = voyageai.VOYAGE_EMBED_DEFAULT_MODEL,
    input_type: Optional[str] = None,
    **kwargs,
) -> List[List[float]]:
    """Get Voyage embedding for a list of text strings (async).

    Args:
        list_of_text (list): A list of text strings to embed.
        model (str): Name of the model to use.
        input_type (str): Type of the input text. Defalut to None, meaning the type is unspecified.
            Other options include: "query", "document".
    """
    warnings.warn(
        "This function will be deprecated. We recommend using the `embed()` function in `voyageai.AsyncClient`."
    )
    if len(list_of_text) <= MAX_BATCH_SIZE:
        return await _aget_embeddings(list_of_text, model, input_type, **kwargs)

    assert (
        len(list_of_text) <= MAX_LIST_LENGTH
    ), f"The length of list_of_text should not be larger than {MAX_LIST_LENGTH}."

    semaphore = asyncio.Semaphore(value=DEFAULT_CONCURRENCE)
    rate_limit = AsyncLimiter(DEFAULT_RPM, 60)

    batches = [
        list_of_text[i : i + MAX_BATCH_SIZE]
        for i in range(0, len(list_of_text), MAX_BATCH_SIZE)
    ]
    async_tasks = [
        _aget_embeddings(
            batch, model, input_type, semaphore=semaphore, rate_limit=rate_limit
        )
        for batch in batches
    ]
    results = await asyncio.gather(*async_tasks)
    return sum(results, [])



from voyageai.api_resources import APIResource


class Reranking(APIResource):

    OBJECT_NAME = "rerank"

    @classmethod
    def create(cls, *args, **kwargs):
        """
        Creates a new reranking for the provided input and parameters.
        """
        response = super().create(*args, **kwargs)
        return response

    @classmethod
    async def acreate(cls, *args, **kwargs):
        """
        Creates a new reranking for the provided input and parameters.
        """
        response = await super().acreate(*args, **kwargs)
        return response



from voyageai.api_resources.response import VoyageResponse
from voyageai.api_resources.api_requestor import VoyageHttpResponse

from voyageai.api_resources.api_resource import APIResource
from voyageai.api_resources.embedding import Embedding
from voyageai.api_resources.reranking import Reranking
